import json
import os
import shutil
from concurrent.futures import ProcessPoolExecutor
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import Literal, Union

import ipywidgets as widgets
import matplotlib.pyplot as plt
import numpy as np
import pyvista as pv
import torch
import torch.utils
import torch.utils.data
from ipywidgets import fixed, interact
from matplotlib import cm
from tensordict import TensorDict
from tqdm import tqdm

from cooldata.pyvista_flow_field_dataset import PyvistaFlowFieldDataset, PyvistaSample

VoxelField = Literal[
    "Pressure",
    "Temperature",
    "Velocity",
    "Position",
    "TurbulentKineticEnergy",
    "TurbulentDissipationRate",
]
voxel_fields: list[VoxelField] = [
    "Pressure",
    "Temperature",
    "Velocity",
    "Position",
    "TurbulentKineticEnergy",
    "TurbulentDissipationRate",
]

Normalization = dict[
    VoxelField, Union[tuple[list[float], list[float]], tuple[float, float]]
]
"""
A dictionary that maps field names to tuples of (mean, std) for normalization.
"""


class VoxelFlowFieldSample:
    def __init__(
        self,
        path: str,
        bounding_box: tuple[float, float, float, float, float, float],
        resolution: tuple[int, int, int],
        normalization: Normalization | None = None,
    ):
        """A sample of a voxelized flow field. The data is stored in a file in the TensorDict format on disk in an unnormalized form. A normalization can be applied dynamically when getting data.

        Args:
        - path: Path to the file containing the voxelized flow field.
        - bounding_box: Tuple (xmin, xmax, ymin, ymax, zmin, zmax) defining the bounding box of the flow field.
        - resolution: Tuple (nx, ny, nz) defining the resolution of the voxel grid.
        - normalization: A normalization that is applied dynamically when getting data. A dictionary that maps field names to tuples of (mean, std) for normalization. The
        """

        self.path = path
        self.bounding_box = bounding_box
        self.resolution = resolution
        self._data: TensorDict | None = None
        self.normalization = normalization

    @property
    def data(self):
        if self._data is None:
            self._data = TensorDict.load(self.path)
            # TODO: Check if data is a valid flow field dataset, i.e., has the necessary point data
        return self._data

    def get_field(self, field: VoxelField, normalized=True) -> torch.Tensor:
        """
        Returns the specified field of the flow field as a torch.Tensor.
        Shape:
        - Pressure, Temperature: (resolution_x, resolution_y, resolution_z)
        - Velocity, Position: (resolution_x, resolution_y, resolution_z, 3)
        """
        if self.normalization is not None and normalized:
            mean, std = self.normalization[field]
            return (
                self.data[field] - torch.tensor(mean, device=self.data.device)
            ) / torch.tensor(std, device=self.data.device)
        return self.data[field]

    @property
    def Y(self):
        return torch.cat(
            [
                self.get_field("Velocity"),
                self.get_field("Pressure").unsqueeze(-1),
                self.get_field("Temperature").unsqueeze(-1),
            ],
            dim=-1,
        )

    @property
    def mask(self):
        return self.data["Mask"]

    @classmethod
    def from_mask_y(
        cls,
        mask: torch.Tensor,
        Y: torch.Tensor,
        bounding_box: tuple[float, float, float, float, float, float],
        resolution: tuple[int, int, int],
        save_path: str,
        normalization: Normalization | None = None,
    ) -> "VoxelFlowFieldSample":
        assert mask.shape == tuple(resolution), (
            f"Mask shape {mask.shape} does not match resolution {resolution}"
        )
        assert Y.shape == (resolution[0], resolution[1], resolution[2], 5), (
            f"Y shape {Y.shape} does not match resolution {resolution}"
        )
        # check if the mask is a boolean tensor
        assert mask.dtype == torch.bool, f"Mask dtype {mask.dtype} is not boolean"
        assert Y.dtype == torch.float32, f"Y dtype {Y.dtype} is not float32"
        velocity = Y[:, :, :, :3]
        pressure = Y[:, :, :, 3]
        temperature = Y[:, :, :, 4]
        # get position from bounding box and resolution
        xmin, xmax, ymin, ymax, zmin, zmax = bounding_box
        x = np.linspace(xmin, xmax, resolution[0])
        y = np.linspace(ymin, ymax, resolution[1])
        z = np.linspace(zmin, zmax, resolution[2])
        x, y, z = np.meshgrid(x, y, z, indexing="ij")
        position = torch.tensor(np.stack([x, y, z], axis=-1), dtype=torch.float32)
        data = TensorDict(
            {
                "Pressure": pressure,
                "Temperature": temperature,
                "Velocity": velocity,
                "Mask": mask,
                "Position": position,
            }
        )
        data = data.to("cpu")
        # denormalize the data
        if normalization is not None:
            for field in data.keys():
                if field in normalization:
                    mean, std = normalization[field]
                    data[field] = data[field] * torch.tensor(
                        std, device=data.device
                    ) + torch.tensor(mean, device=data.device)
        data.save(save_path)
        return cls(save_path, bounding_box, resolution, normalization)

    @classmethod
    def from_pyvista(
        cls,
        sample: PyvistaSample,
        save_path: str,
        resolution: tuple[int, int, int],
        bounding_box: tuple[float, float, float, float, float, float],
    ) -> "VoxelFlowFieldSample":
        """
        Interpolates the volume data from the sample to a voxel grid and saves it to a file.
        """
        if os.path.exists(save_path):
            return cls(save_path, bounding_box, resolution)
        xmin, xmax, ymin, ymax, zmin, zmax = bounding_box
        x = np.linspace(xmin, xmax, resolution[0])
        y = np.linspace(ymin, ymax, resolution[1])
        z = np.linspace(zmin, zmax, resolution[2])
        x, y, z = np.meshgrid(x, y, z, indexing="ij")
        grid = pv.StructuredGrid(x, y, z)
        was_loaded = sample.is_loaded
        volume_data = sample.volume_data[0][0][0]
        interpolated = grid.sample(volume_data)
        velocities = [interpolated[f"Velocity_{i}"] for i in range(3)]
        data = TensorDict(
            {
                "Pressure": torch.tensor(
                    interpolated["Pressure"].reshape(resolution, order="F"),
                    dtype=torch.float32,
                ),
                "Temperature": torch.tensor(
                    interpolated["Temperature"].reshape(resolution, order="F"),
                    dtype=torch.float32,
                ),
                "Velocity": torch.stack(
                    [
                        torch.tensor(
                            v.reshape(resolution, order="F"), dtype=torch.float32
                        )
                        for v in velocities
                    ],
                    dim=-1,
                ),
                "Mask": torch.tensor(
                    interpolated["vtkValidPointMask"].reshape(resolution, order="F"),
                    dtype=torch.bool,
                ),
                "Position": torch.tensor(
                    np.stack([x, y, z], axis=-1).reshape(resolution + (3,), order="F"),
                    dtype=torch.float32,
                ),
                "TurbulentKineticEnergy": torch.tensor(
                    interpolated["TurbulentKineticEnergy"].reshape(
                        resolution, order="F"
                    ),
                    dtype=torch.float32,
                ),
                "TurbulentDissipationRate": torch.tensor(
                    interpolated["TurbulentDissipationRate"].reshape(
                        resolution, order="F"
                    ),
                    dtype=torch.float32,
                ),
            }
        )
        data.save(save_path)
        if not was_loaded:
            sample.unload()
        return cls(save_path, bounding_box, resolution)

    def load(self):
        self._data
        return self

    def unload(self):
        self._data = None
        return self

    def to_pyvista(self):
        xmin, xmax, ymin, ymax, zmin, zmax = self.bounding_box
        x, y, z = np.mgrid[
            xmin : xmax : complex(self.resolution[0]),
            ymin : ymax : complex(self.resolution[1]),
            zmin : zmax : complex(self.resolution[2]),
        ]
        grid = pv.StructuredGrid(x, y, z)
        grid["Pressure"] = self.data["Pressure"].numpy().flatten(order="F")
        grid["Temperature"] = self.data["Temperature"].numpy().flatten(order="F")
        grid["Velocity"] = self.data["Velocity"].numpy().reshape(-1, 3, order="F")
        grid["vtkValidPointMask"] = (
            self.data["Mask"].numpy().flatten(order="F").astype(np.int8)
        )
        grid["vtkGhostType"] = np.zeros(len(grid.points), dtype=np.uint8)
        grid["vtkGhostType"][~self.data["Mask"].numpy().flatten(order="F")] = 32
        grid.cell_data["vtkGhostType"] = np.zeros(grid.n_cells, dtype=np.uint8)
        grid.cell_data["vtkGhostType"][
            ~self.data["Mask"].numpy()[1:, 1:, 1:].flatten(order="F")
        ] = 32
        return grid

    def plot(self, field: VoxelField):
        grid = self.to_pyvista()
        grid.plot(scalars=field, cmap="viridis")

    def plot_slice(
        self,
        field: VoxelField,
        slice_idx: int | None = None,
        axis: Literal["x", "y", "z"] = "z",
    ):
        """
        Plots a slice of the field at the specified index along the specified axis.
        Args:
        - field: The field to plot.
        - slice_idx: The index of the slice to plot. If None, the middle slice is used.
        - axis: The axis along which to plot the slice. Can be "x", "y", or "z".
        """
        field_values = self.get_field(field, normalized=False)
        field_np = field_values.cpu().numpy()
        slice: np.ndarray | None = None
        if axis == "x":
            if slice_idx is None:
                slice_idx = field_np.shape[0] // 2
            slice = field_np[slice_idx, :, :]
        elif axis == "y":
            if slice_idx is None:
                slice_idx = field_np.shape[1] // 2
            slice = field_np[:, slice_idx, :]
        elif axis == "z":
            if slice_idx is None:
                slice_idx = field_np.shape[2] // 2
            slice = field_np[:, :, slice_idx]
        else:
            raise ValueError("Axis must be 'x', 'y', or 'z'.")
        title_addon = ""
        if slice.ndim == 3:
            slice = np.sqrt(
                np.sum(slice**2, axis=-1)
            )  # If it's a vector field, take the magnitude
            title_addon = " (magnitude)"
        plt.imshow(slice, cmap="viridis")
        plt.colorbar()
        if axis == "x":
            plt.xlabel("y")
            plt.ylabel("z")
            plt.yticks(
                ticks=np.arange(0, self.resolution[1], step=5),
                labels=[
                    f"{val:.2f}"
                    for val in np.linspace(
                        self.bounding_box[2],
                        self.bounding_box[3],
                        num=self.resolution[1],
                    )[::5]
                ],
            )
            plt.xticks(
                ticks=np.arange(0, self.resolution[2], step=5),
                labels=[
                    f"{val:.2f}"
                    for val in np.linspace(
                        self.bounding_box[4],
                        self.bounding_box[5],
                        num=self.resolution[2],
                    )[::5]
                ],
            )
            plt.title(
                f"{field} at x = {self.bounding_box[0] + (slice_idx / self.resolution[0]) * (self.bounding_box[1] - self.bounding_box[0]):.2f}{title_addon}"
            )
        elif axis == "y":
            plt.xlabel("x")
            plt.ylabel("z")
            plt.yticks(
                ticks=np.arange(0, self.resolution[0], step=5),
                labels=[
                    f"{val:.2f}"
                    for val in np.linspace(
                        self.bounding_box[0],
                        self.bounding_box[1],
                        num=self.resolution[0],
                    )[::5]
                ],
            )
            plt.xticks(
                ticks=np.arange(0, self.resolution[2], step=5),
                labels=[
                    f"{val:.2f}"
                    for val in np.linspace(
                        self.bounding_box[4],
                        self.bounding_box[5],
                        num=self.resolution[2],
                    )[::5]
                ],
            )
            plt.title(
                f"{field} at y = {self.bounding_box[2] + (slice_idx / self.resolution[1]) * (self.bounding_box[3] - self.bounding_box[2]):.2f}{title_addon}"
            )
        elif axis == "z":
            plt.xlabel("x")
            plt.ylabel("y")
            plt.yticks(
                ticks=np.arange(0, self.resolution[0], step=5),
                labels=[
                    f"{val:.2f}"
                    for val in np.linspace(
                        self.bounding_box[0],
                        self.bounding_box[1],
                        num=self.resolution[0],
                    )[::5]
                ],
            )
            plt.xticks(
                ticks=np.arange(0, self.resolution[1], step=5),
                labels=[
                    f"{val:.2f}"
                    for val in np.linspace(
                        self.bounding_box[2],
                        self.bounding_box[3],
                        num=self.resolution[1],
                    )[::5]
                ],
            )
            plt.title(
                f"{field} at z = {self.bounding_box[4] + (slice_idx / self.resolution[2]) * (self.bounding_box[5] - self.bounding_box[4]):.2f}{title_addon}"
            )

        # preserve aspect ratio
        plt.gca().set_aspect("equal", adjustable="box")
        plt.grid()
        plt.tight_layout()
        plt.show()

    def plot_slice_interactively(
        self, field: VoxelField, axis: Literal["x", "y", "z"] = "z"
    ):
        """
        Plots a slice of the field at the specified index along the specified axis.
        Args:
        - field: The field to plot.
        - axis: The axis along which to plot the slice. Can be "x", "y", or "z".
        """

        def update_plot(slice_idx):
            self.plot_slice(field, slice_idx, axis)

        interact(
            update_plot,
            slice_idx=widgets.IntSlider(
                min=0,
                max=self.resolution[{"x": 0, "y": 1, "z": 2}[axis]] - 1,
                step=1,
                value=self.resolution[{"x": 0, "y": 1, "z": 2}[axis]] // 2,
            ),
            field=fixed(field),
            axis=fixed(axis),
        )


@dataclass
class VoxelFlowFieldDatasetConfig:
    """Configuration for creating a VoxelFlowFieldDataset from a PyvistaFlowFieldDataset."""

    pyvista_dataset: PyvistaFlowFieldDataset
    resolution: tuple[int, int, int] = (32, 32, 32)


class VoxelFlowFieldDataset(torch.utils.data.Dataset):
    def __init__(
        self,
        cache_dir: str,
        config: VoxelFlowFieldDatasetConfig | None = None,
        resume_loading_from_cache: bool = True,
    ):
        """
        Dataset of voxelized flow fields. The constructor either loads the dataset from a cache directory or converts a
        PyvistaFlowFieldDataset to a DGLFlowFieldDataset.
        """
        self.cache_dir = Path(os.path.abspath(cache_dir))
        if not self.cache_dir.exists():
            self.cache_dir.mkdir(parents=True)
        self.samples: list[VoxelFlowFieldSample] = []
        if config is not None:
            # clear the cache directory
            if not resume_loading_from_cache and self.cache_dir.exists():
                shutil.rmtree(self.cache_dir)
            self.cache_dir.mkdir(parents=True, exist_ok=True)
            config.pyvista_dataset.unload()
            self.resolution = config.resolution
            if os.path.exists(os.path.join(self.cache_dir, "metadata.json")):
                metadata = json.load(
                    open(os.path.join(self.cache_dir, "metadata.json"))
                )
                if tuple(metadata["resolution"]) == config.resolution:
                    # If the metadata matches, we can use it
                    self.resolution = metadata["resolution"]
                    self.bounding_box = metadata["bounding_box"]
                    self.normalization = metadata["normalization"]
            if not hasattr(self, "bounding_box"):
                # If the bounding box is not set, we use the bounding box of the PyvistaFlowFieldDataset
                self.bounding_box = config.pyvista_dataset.get_bounds()

            args_list = [
                (
                    config.pyvista_dataset[i],
                    self.cache_dir,
                    i,
                    config.resolution,
                    self.bounding_box,
                )
                for i in range(len(config.pyvista_dataset))
            ]
            with ProcessPoolExecutor() as executor:
                results = list(
                    tqdm(
                        executor.map(_create_voxel_sample, args_list),
                        total=len(args_list),
                        desc="Voxelizing samples",
                    )
                )
            self.samples.extend(results)
            if not hasattr(self, "normalization"):
                self.normalization = self.compute_normalization()

            # else, we compute the normalization and save the metadata
            json.dump(
                {
                    "resolution": config.resolution,
                    "bounding_box": self.bounding_box,
                    "normalization": self.normalization,
                },
                open(os.path.join(self.cache_dir, "metadata.json"), "w"),
            )
        else:
            metadata = json.load(open(os.path.join(self.cache_dir, "metadata.json")))
            self.resolution = metadata["resolution"]
            self.bounding_box = metadata["bounding_box"]
            self.normalization = metadata["normalization"]
            for file in os.listdir(self.cache_dir):
                if file.endswith(".pt"):
                    self.samples.append(
                        VoxelFlowFieldSample(
                            os.path.join(self.cache_dir, file),
                            self.bounding_box,
                            self.resolution,
                        )
                    )

    def __len__(self):
        return len(self.samples)

    def __getitem__(self, idx: int | slice | list[int]):
        if isinstance(idx, slice):
            return self.slice(idx.start, idx.stop)
        if isinstance(idx, list):
            new_ds = VoxelFlowFieldDataset(self.cache_dir)
            new_ds.samples = [self.samples[i] for i in idx]
            new_ds.bounding_box = self.bounding_box
            new_ds.resolution = self.resolution
            new_ds.normalization = self.normalization
            return new_ds
        if idx < 0 or idx >= len(self.samples):
            raise IndexError(
                f"Index {idx} out of bounds for dataset of length {len(self.samples)}"
            )
        if not isinstance(idx, int):
            raise TypeError(f"Index must be an integer or a slice, got {type(idx)}")
        return self.samples[idx]

    def slice(self, start: int, end: int):
        """
        Returns a slice of the dataset.

        Args:
        - start: The start index of the slice.
        - end: The end index of the slice.

        Returns:
        A new VoxelFlowFieldDataset containing the specified slice.
        """
        new_ds = VoxelFlowFieldDataset(self.cache_dir)
        new_ds.samples = self.samples[start:end]
        new_ds.bounding_box = self.bounding_box
        new_ds.resolution = self.resolution
        new_ds.normalization = self.normalization
        return new_ds

    def shuffle(self):
        """Shuffles the dataset in place."""
        np.random.shuffle(self.samples)
        return self

    def compute_normalization(self) -> Normalization:
        normalization: Normalization = {}
        for field in voxel_fields:
            sample_means = []
            sample_stds = []
            for sample in self.samples:
                sample_means.append(sample.get_field(field).mean(dim=(0, 1, 2)))
                sample_stds.append(sample.get_field(field).std(dim=(0, 1, 2)))
            mean = torch.stack(sample_means).mean(dim=0).tolist()
            std = torch.stack(sample_stds).mean(dim=0).tolist()
            normalization[field] = (mean, std)
        return normalization

    def normalize(self):
        """Normalizes the dataset in place."""
        for sample in self.samples:
            sample.normalization = self.normalization
        return self

    def prediction_to_sample(self, mask: torch.Tensor, Y: torch.Tensor):
        """
        Converts a prediction to a sample. The mask is used to create a new sample with the same bounding box and resolution as the original dataset.

        Args:
        - mask: The mask of the prediction. Shape: (resolution_x, resolution_y, resolution_z)
        - Y: The prediction. Shape: (resolution_x, resolution_y, resolution_z, 5)
        """
        # get the bounding box and resolution from the original dataset
        bounding_box = self.bounding_box
        resolution = self.resolution
        now = datetime.now()

        # create a new sample from the prediction
        timestamp_formatted = now.strftime("%Y-%m-%d %H:%M:%S.%f")[:-3]
        return VoxelFlowFieldSample.from_mask_y(
            mask,
            Y,
            bounding_box,
            resolution,
            os.path.join(
                self.cache_dir, "predictions", f"prediction_{timestamp_formatted}.pt"
            ),
            normalization=self.normalization,
        )

    def unnormalize(self):
        """Unnormalizes the dataset in place."""
        for sample in self.samples:
            sample.normalization = None
        return self

    def get_default_loadable_dataset(self):
        """Get a dataset that returns the mask as X and all flow features concatenated as Y"""
        return DefaultVoxelDataset(self)


class DefaultVoxelDataset(torch.utils.data.Dataset):
    def __init__(self, ds: VoxelFlowFieldDataset):
        super().__init__()
        self.ds = ds

    def __getitem__(self, index: int):
        item = self.ds[index]
        return item.mask, item.Y

    def __len__(self):
        return len(self.ds)


def _create_voxel_sample(args):
    sample_pv, cache_dir, i, resolution, bounding_box = args
    return VoxelFlowFieldSample.from_pyvista(
        sample_pv,
        os.path.join(cache_dir, f"{i}.pt"),
        resolution,
        bounding_box,
    )



import json
import os
from pathlib import Path
from typing import Callable, List, Optional

try:
    import dgl
except ImportError as e:
    raise ImportError(
        f"""DGL is not installed. Please install it manually by following the instructions at: 
        "https://www.dgl.ai/pages/start.html
        {e}"""
    )

import concurrent.futures
import shutil

import numpy as np
import pyvista as pv
import torch
from tqdm import tqdm

from cooldata.pyvista_flow_field_dataset import (
    PyvistaFlowFieldDataset,
    PyvistaSample,
    SurfaceFieldType,
)


def process_sample(args):
    cache_dir, sample, i = args
    g = DGLVolumeFlowFieldDataset.pyvista_to_volume_dgl(sample)
    print("Saving graph to cache directory:", os.path.join(cache_dir, f"{i}.dgl"))
    dgl.save_graphs(os.path.join(cache_dir, f"{i}.dgl"), g)
    print("Graph saved successfully.")


class DGLVolumeFlowFieldDataset(torch.utils.data.Dataset):
    def __init__(
        self, cache_dir: str, pyvista_dataset: PyvistaFlowFieldDataset | None = None, parallel_conversion: bool = True
    ):
        """
        Creates a new DGLVolumeFlowFieldDataset. If a PyvistaFlowFieldDataset is provided, it will be converted to DGLGraphs and stored in the cache directory. If not, the dataset will be loaded from the cache directory.
        Parameters:
        -----------
        cache_dir: str
            The directory where the dataset converted to DGLGraphs is stored.
        polydata: pv.PolyData
            The directory where the dataset converted to DGLGraphs is stored. Default None.
        """
        self.cache_dir = Path(os.path.abspath(cache_dir))
        self.cache_dir.mkdir(parents=True, exist_ok=True)
        self.node_stats: tuple[dict, dict] | None = None
        self.edge_stats: tuple[dict, dict] | None = None
        if not self.cache_dir.exists():
            self.cache_dir.mkdir(parents=True)
        existing_files = [f for f in self.cache_dir.iterdir() if f.suffix == ".dgl"]
        if len(existing_files) == len(pyvista_dataset or []):
            self.files = existing_files
            return
        if pyvista_dataset is not None:
            # clear the cache directory
            shutil.rmtree(self.cache_dir)
            self.cache_dir.mkdir(parents=True)
            pyvista_dataset.unload()
            if parallel_conversion:
                args = [
                    (self.cache_dir, pyvista_dataset[i], i)
                    for i in range(len(pyvista_dataset))
                ]
                with concurrent.futures.ProcessPoolExecutor() as executor:
                    list(
                        tqdm(
                            executor.map(process_sample, args),
                            total=len(pyvista_dataset),
                            desc="Converting Pyvista dataset to DGLGraphs",
                        )
                    )
            else:
                for i in tqdm(
                    range(len(pyvista_dataset)),
                    desc="Converting Pyvista dataset to DGLGraphs",
                ):
                    process_sample((self.cache_dir, pyvista_dataset[i], i))
        self.files = [f for f in self.cache_dir.iterdir() if f.suffix == ".dgl"]

    def __len__(self):
        return len(self.files)

    def __getitem__(self, idx):
        filename = os.path.join(self.cache_dir, self.files[idx])
        if filename.endswith(".dgl"):
            g = dgl.load_graphs(filename)[0][0]
            if self.node_stats is not None and self.edge_stats is not None:
                self.normalize_inplace(g)
            return g
        else:
            raise ValueError(f"File {filename} is not a DGLGraph file.")

    @classmethod
    def pyvista_to_volume_dgl(cls, sample: PyvistaSample) -> dgl.DGLGraph:
        """
        Convert a Pyvista UnstructuredGrid object to a DGLGraph of the volume flow field.

        Parameters:
        -----------
        grid: pv.UnstructuredGrid
            The UnstructuredGrid object to convert.

        Returns:
        --------
        DGLGraph: The converted graph.
        """

        grid = sample.volume_data[0][0][0]
        edges_from = []
        edges_to = []

        # TODO: Speed up this loop
        for i in tqdm(range(grid.n_cells)):
            neighbors = grid.cell_neighbors(i)
            edges_from.extend([i] * len(neighbors))
            edges_to.extend(neighbors)

        graph = dgl.graph((edges_from, edges_to), num_nodes=grid.n_cells)
        # TODO add the attributes to the nodes from the grid
        graph.ndata["Pressure"] = torch.tensor(
            grid.cell_data["Pressure"], dtype=torch.float32
        )
        graph.ndata["Temperature"] = torch.tensor(
            grid.cell_data["Temperature"], dtype=torch.float32
        )
        centers = torch.tensor(grid.cell_centers().points, dtype=torch.float32)
        graph.ndata["Position"] = centers
        velocities = torch.stack(
            [
                torch.tensor(grid.cell_data["Velocity_0"], dtype=torch.float32),
                torch.tensor(grid.cell_data["Velocity_1"], dtype=torch.float32),
                torch.tensor(grid.cell_data["Velocity_2"], dtype=torch.float32),
            ],
            dim=1,
        )
        graph.ndata["Velocity"] = velocities
        graph.ndata["TurbulentDissipationRate"] = torch.tensor(
            grid.cell_data["TurbulentDissipationRate"], dtype=torch.float32
        )
        graph.ndata["TurbulentKineticEnergy"] = torch.tensor(
            grid.cell_data["TurbulentKineticEnergy"], dtype=torch.float32
        )
        graph.ndata["Volume"] = torch.tensor(
            grid.cell_data["Volume"], dtype=torch.float32
        )
        connectivity_vectors = (
            graph.ndata["Position"][edges_to] - graph.ndata["Position"][edges_from]
        )
        graph.edata["dx"] = connectivity_vectors
        return graph

    def volume_dgl_to_pv(self, graph: dgl.DGLGraph) -> pv.PolyData:
        """
        Convert a DGLGraph of the volume flow field to a Pyvista PolyData object.

        Parameters:
        -----------
        graph: dgl.DGLGraph
            The DGLGraph to convert.

        Returns:
        --------
        pv.PolyData: The converted PolyData object.
        """
        raise NotImplementedError("Implement this method")

    def normalize(self):
        if os.path.exists(os.path.join(self.cache_dir, "stats.json")):
            with open(os.path.join(self.cache_dir, "stats.json"), "r") as f:
                stats = json.load(f)
                if "node_stats" in stats and "edge_stats" in stats:
                    self.node_stats = stats["node_stats"]
                    self.edge_stats = stats["edge_stats"]
        else:
            self.node_stats = self.compute_node_stats()
            self.edge_stats = self.compute_edge_stats()
            with open(os.path.join(self.cache_dir, "stats.json"), "w") as f:
                json.dump(
                    {
                        "node_stats": self.node_stats,
                        "edge_stats": self.edge_stats,
                    },
                    f,
                )

    def denormalize(self):
        self.node_stats = None
        self.edge_stats = None

    def normalize_inplace(self, graph: dgl.DGLGraph):
        """
        Normalize the features of the graph.

        Parameters:
        -----------
        graph: dgl.DGLGraph
            The graph to normalize.
        """
        for key in graph.ndata.keys():
            if key not in self.node_stats[0] or key not in self.node_stats[1]:
                continue
            if graph.ndata[key].dtype != torch.float32:
                continue
            graph.ndata[key] = (
                graph.ndata[key]
                - torch.tensor(self.node_stats[0][key], device=graph.device)
            ) / torch.tensor(self.node_stats[1][key], device=graph.device)
        for key in graph.edata.keys():
            if key not in self.edge_stats[0] or key not in self.edge_stats[1]:
                continue
            if graph.edata[key].dtype != torch.float32:
                continue
            graph.edata[key] = (
                graph.edata[key]
                - torch.tensor(self.edge_stats[0][key], device=graph.device)
            ) / torch.tensor(self.edge_stats[1][key], device=graph.device)

    def denormalize_inplace(self, graph: dgl.DGLGraph):
        """
        Denormalize the features of the graph.

        Parameters:
        -----------
        graph: dgl.DGLGraph
            The graph to denormalize.
        """
        for key in graph.ndata.keys():
            if key not in self.node_means or key not in self.node_stds:
                continue
            if graph.ndata[key].dtype != torch.float32:
                continue
            graph.ndata[key] = (
                graph.ndata[key]
                * torch.tensor(self.node_stds[key], device=graph.device)
            ) + torch.tensor(self.node_means[key], device=graph.device)
        for key in graph.edata.keys():
            if key not in self.edge_means or key not in self.edge_stds:
                continue
            if graph.edata[key].dtype != torch.float32:
                continue
            graph.edata[key] = (
                graph.edata[key]
                * torch.tensor(self.edge_stds[key], device=graph.device)
            ) + torch.tensor(self.edge_means[key], device=graph.device)

    def compute_node_stats(self) -> tuple[dict[str, float], dict[str, float]]:
        """
        Compute the mean and standard deviation of the node features.
        """
        graph_means: dict[str, list[float]] = {}
        graph_stds: dict[str, list[float]] = {}
        for i in range(len(self)):
            graph = self[i]
            for key in graph.ndata.keys():
                if graph.ndata[key].dtype != torch.float32:
                    continue
                if key not in graph_means:
                    graph_means[key] = []
                    graph_stds[key] = []
                graph_means[key].append(graph.ndata[key].mean(dim=0).tolist())
                graph_stds[key].append(graph.ndata[key].std(dim=0).tolist())
        means = {
            key: (np.sum(graph_means[key], axis=0) / len(graph_means[key])).tolist()
            for key in graph_means
        }
        stds = {
            key: (
                np.sqrt(
                    np.sum(np.square(graph_stds[key]), axis=0) / len(graph_stds[key])
                )
            ).tolist()
            for key in graph_stds
        }
        return means, stds

    def compute_edge_stats(self) -> tuple[dict[str, float], dict[str, float]]:
        """
        Compute the mean and standard deviation of the edge features.
        """
        graph_means: dict[str, list[float]] = {}
        graph_stds: dict[str, list[float]] = {}
        for i in range(len(self)):
            graph = self[i]
            for key in graph.edata.keys():
                if graph.edata[key].dtype != torch.float32:
                    continue
                if key not in graph_means:
                    graph_means[key] = []
                    graph_stds[key] = []
                graph_means[key].append(graph.edata[key].mean(dim=0).tolist())
                graph_stds[key].append(graph.edata[key].std(dim=0).tolist())
        means = {
            key: (np.sum(graph_means[key], axis=0) / len(graph_means[key])).tolist()
            for key in graph_means
        }
        stds = {
            key: (
                np.sqrt(
                    np.sum(np.square(graph_stds[key]), axis=0) / len(graph_stds[key])
                )
            ).tolist()
            for key in graph_stds
        }
        return means, stds

    @classmethod
    def l2_loss(cls, graph1: dgl.DGLGraph, graph2: dgl.DGLGraph):
        """
        Compute the L2 loss between two DGLGraphs.

        Parameters:
        -----------
        graph1: dgl.DGLGraph
            The first graph.
        graph2: dgl.DGLGraph
            The second graph.

        Returns:
        --------
        float: The L2 loss between the two graphs.
        """
        raise NotImplementedError("Implement this method")


def process_surface_sample(args):
    cache_dir, sample, i, patches_to_include = args
    g = DGLSurfaceFlowFieldDataset.pyvista_to_surface_dgl(sample, patches_to_include)
    dgl.save_graphs(os.path.join(cache_dir, f"{i}.dgl"), g)


class DGLSurfaceFlowFieldDataset(torch.utils.data.Dataset):
    def __init__(
        self,
        cache_dir: str,
        pyvista_dataset: PyvistaFlowFieldDataset | None = None,
        normalize: bool = True,
        patches_to_include: Optional[List[int]] = None,
    ):
        """Dataset of surface flow fields represented as DGLGraphs.

        If a PyvistaFlowFieldDataset is provided, it will be converted to DGLGraphs and stored in the cache directory. If not, the dataset will be loaded from the cache directory.

        Parameters:
        -----------
        cache_dir: str
            The directory where the dataset converted to DGLGraphs is stored.
        polydata: pv.PolyData
            The directory where the dataset converted to DGLGraphs is stored. Default None.
        """
        self.cache_dir = Path(os.path.abspath(cache_dir))
        if not self.cache_dir.exists():
            os.makedirs(self.cache_dir)
        existing = [f for f in os.listdir(self.cache_dir) if f.endswith(".dgl")]
        if pyvista_dataset is not None and len(existing) != len(pyvista_dataset):
            # clear the cache directory
            for f in os.listdir(self.cache_dir):
                os.remove(os.path.join(self.cache_dir, f))
            args = [
                (self.cache_dir, pyvista_dataset[i], i, patches_to_include)
                for i in range(len(pyvista_dataset))
            ]
            with concurrent.futures.ProcessPoolExecutor() as executor:
                list(
                    tqdm(
                        executor.map(process_surface_sample, args),
                        total=len(pyvista_dataset),
                        desc="Converting Pyvista dataset to DGLGraphs (surface)",
                    )
                )
        self.files = [f for f in os.listdir(self.cache_dir) if f.endswith(".dgl")]

        # first do not normalize to compute the stats, then normalize
        self.do_normalization = False
        if pyvista_dataset is not None:
            node_stats = self.compute_node_stats()
            edge_stats = self.compute_edge_stats()
            with open(os.path.join(self.cache_dir, "stats.json"), "w") as f:
                json.dump({"node_stats": node_stats, "edge_stats": edge_stats}, f)
        with open(os.path.join(self.cache_dir, "stats.json"), "r") as f:
            stats = json.load(f)
            node_stats: tuple[dict[str, float], dict[str, float]] = stats["node_stats"]
            edge_stats: tuple[dict[str, float], dict[str, float]] = stats["edge_stats"]
            self.node_means, self.node_stds = node_stats
            self.edge_means, self.edge_stds = edge_stats
        self.do_normalization = normalize

    def __len__(self):
        return len(self.files)

    def __getitem__(self, idx):
        """
        Get the graph at the given index."""
        filename = os.path.join(self.cache_dir, self.files[idx])
        g = dgl.load_graphs(filename)[0][0]
        if self.do_normalization:
            g = self.normalize(g)
        return g

    def slice(self, start: int, end: int) -> "DGLSurfaceFlowFieldDataset":
        """
        Get a slice of the dataset.
        """
        new_dataset = DGLSurfaceFlowFieldDataset(
            cache_dir=self.cache_dir,
            pyvista_dataset=None,
            normalize=self.do_normalization,
        )
        new_dataset.files = self.files[start:end]
        new_dataset.node_means = self.node_means
        new_dataset.node_stds = self.node_stds
        new_dataset.edge_means = self.edge_means
        new_dataset.edge_stds = self.edge_stds
        return new_dataset

    def shuffle(self) -> None:
        """
        Shuffle the dataset.
        """
        np.random.shuffle(self.files)

    def select_subset(self, indices: List[int]) -> "DGLSurfaceFlowFieldDataset":
        """
        Select a subset of the dataset based on the given indices.
        """
        new_dataset = DGLSurfaceFlowFieldDataset(
            cache_dir=self.cache_dir,
            pyvista_dataset=None,
            normalize=self.do_normalization,
        )
        new_dataset.files = [self.files[i] for i in indices]
        new_dataset.node_means = self.node_means
        new_dataset.node_stds = self.node_stds
        new_dataset.edge_means = self.edge_means
        new_dataset.edge_stds = self.edge_stds
        return new_dataset

    @classmethod
    def pyvista_to_surface_dgl(
        cls, sample: PyvistaSample, block_indices: Optional[List[int]] = None
    ) -> dgl.DGLGraph:
        """
        Convert a Pyvista UnstructuredGrid object to a DGLGraph of the volume flow field.

        Parameters:
        -----------
        grid: pv.UnstructuredGrid
            The UnstructuredGrid object to convert.

        Returns:
        --------
        DGLGraph: The converted graph.
        """
        if block_indices is None:
            block_indices = list(range(len(sample.surface_data[0])))

        for block_index in block_indices:
            grid = sample.surface_data[0][block_index]
            # TODO: Speed up this loop
            grid.cell_data["BodyID"] = block_index
            block_name = sample.surface_data[0].get_block_name(block_index)
            surface_type = -1
            if "wall" in block_name.lower():
                surface_type = 0
            elif "inlet" in block_name.lower():
                surface_type = 1
            elif "outlet" in block_name.lower():
                surface_type = 2
            elif "symmetry" in block_name.lower():
                surface_type = 3
            elif "body" in block_name.lower():
                surface_type = 4
            else:
                print(f"Unknown surface type for block {block_index}: {block_name}")
            grid.cell_data["SurfaceType"] = surface_type
            if "WallShearStress_0" not in grid.cell_data:
                grid.cell_data["WallShearStress_0"] = 0.0
                grid.cell_data["WallShearStress_1"] = 0.0
                grid.cell_data["WallShearStress_2"] = 0.0
        combined = sample.surface_data[0].combine(merge_points=True)
        edges_from = []
        edges_to = []
        for i in range(combined.n_cells):
            neighbors = combined.cell_neighbors(i)
            edges_from.extend([i] * len(neighbors))
            edges_to.extend(neighbors)
        graph = dgl.graph((edges_from, edges_to), num_nodes=combined.n_cells)
        graph.ndata["Pressure"] = torch.tensor(
            combined.cell_data["Pressure"], dtype=torch.float32
        )
        graph.ndata["Temperature"] = torch.tensor(
            combined.cell_data["Temperature"], dtype=torch.float32
        )
        centers = torch.tensor(combined.cell_centers().points, dtype=torch.float32)
        graph.ndata["Position"] = centers
        shear_stresses = (
            torch.tensor(combined.cell_data["WallShearStress_0"], dtype=torch.float32),
            torch.tensor(combined.cell_data["WallShearStress_1"], dtype=torch.float32),
            torch.tensor(combined.cell_data["WallShearStress_2"], dtype=torch.float32),
        )

        shear_stress = torch.stack(shear_stresses, dim=1)
        graph.ndata["ShearStress"] = shear_stress
        graph.ndata["Normal"] = torch.tensor(
            combined.extract_surface().face_normals, dtype=torch.float32
        )
        graph.ndata["CellArea"] = torch.tensor(
            combined.compute_cell_sizes(
                length=False, area=True, volume=False
            ).cell_data["Area"],
            dtype=torch.float32,
        )
        graph.ndata["SurfaceType"] = torch.tensor(
            combined.cell_data["SurfaceType"], dtype=torch.int32
        )
        graph.ndata["BodyID"] = torch.tensor(
            combined.cell_data["BodyID"], dtype=torch.int32
        )
        graph.ndata["HeatTransferCoefficient"] = torch.tensor(
            combined.cell_data["HeatTransferCoefficient"], dtype=torch.float32
        )
        connectivity_vectors = (
            graph.ndata["Position"][edges_to] - graph.ndata["Position"][edges_from]
        )
        graph.edata["dx"] = connectivity_vectors
        # graph.ndata['velocity'] = torch.tensor(grid.point_data['Velocity'], dtype=torch.float32)
        # TODO add the attributes to the nodes from the grid
        return graph

    def compute_node_stats(self) -> tuple[dict[str, float], dict[str, float]]:
        """
        Compute the mean and standard deviation of the node features.
        """
        graph_means: dict[str, list[float]] = {}
        graph_stds: dict[str, list[float]] = {}
        for i in range(len(self)):
            graph = self[i]
            for key in graph.ndata.keys():
                if graph.ndata[key].dtype != torch.float32:
                    continue
                if key not in graph_means:
                    graph_means[key] = []
                    graph_stds[key] = []
                graph_means[key].append(graph.ndata[key].mean(dim=0).tolist())
                graph_stds[key].append(graph.ndata[key].std(dim=0).tolist())
        means = {
            key: (np.sum(graph_means[key], axis=0) / len(graph_means[key])).tolist()
            for key in graph_means
        }
        stds = {
            key: (
                np.sqrt(
                    np.sum(np.square(graph_stds[key]), axis=0) / len(graph_stds[key])
                )
            ).tolist()
            for key in graph_stds
        }
        return means, stds

    def compute_edge_stats(self) -> tuple[dict[str, float], dict[str, float]]:
        """
        Compute the mean and standard deviation of the edge features.
        """
        graph_means: dict[str, list[float]] = {}
        graph_stds: dict[str, list[float]] = {}
        for i in range(len(self)):
            graph = self[i]
            for key in graph.edata.keys():
                if graph.edata[key].dtype != torch.float32:
                    continue
                if key not in graph_means:
                    graph_means[key] = []
                    graph_stds[key] = []
                graph_means[key].append(graph.edata[key].mean(dim=0).tolist())
                graph_stds[key].append(graph.edata[key].std(dim=0).tolist())
        means = {
            key: (np.sum(graph_means[key], axis=0) / len(graph_means[key])).tolist()
            for key in graph_means
        }
        stds = {
            key: (
                np.sqrt(
                    np.sum(np.square(graph_stds[key]), axis=0) / len(graph_stds[key])
                )
            ).tolist()
            for key in graph_stds
        }
        return means, stds

    def volume_dgl_to_pyvista(self, graph: dgl.DGLGraph) -> pv.PolyData:
        """
        Convert a DGLGraph of the volume flow field to a Pyvista PolyData object.

        Parameters:
        -----------
        graph: dgl.DGLGraph
            The DGLGraph to convert.

        Returns:
        --------
        pv.PolyData: The converted PolyData object.
        """
        raise NotImplementedError("Implement this method")

    def normalize_inplace(self, graph: dgl.DGLGraph) -> None:
        """
        Normalize the features of the graph in place.

        Parameters:
        -----------
        graph: dgl.DGLGraph
            The graph to normalize.
        """
        for key in graph.ndata.keys():
            if key not in self.node_means or key not in self.node_stds:
                continue
            if graph.ndata[key].dtype != torch.float32:
                continue
            graph.ndata[key] = (
                graph.ndata[key]
                - torch.tensor(self.node_means[key], device=graph.device)
            ) / torch.tensor(self.node_stds[key], device=graph.device)
        for key in graph.edata.keys():
            if key not in self.edge_means or key not in self.edge_stds:
                continue
            if graph.edata[key].dtype != torch.float32:
                continue
            graph.edata[key] = (
                graph.edata[key]
                - torch.tensor(self.edge_means[key], device=graph.device)
            ) / torch.tensor(self.edge_stds[key], device=graph.device)

    def normalize(self, graph: dgl.DGLGraph) -> dgl.DGLGraph:
        """
        Normalize the features of the graph.

        Parameters:
        -----------
        graph: dgl.DGLGraph
            The graph to normalize.

        Returns:
        --------
        dgl.DGLGraph: The normalized graph.
        """
        graph = graph.clone()
        self.normalize_inplace(graph)
        return graph

    def denormalize_inplace(self, graph: dgl.DGLGraph) -> None:
        """
        Denormalize the features of the graph in place.

        Parameters:
        -----------
        graph: dgl.DGLGraph
            The graph to denormalize.
        """

        for key in graph.ndata.keys():
            if key not in self.node_means or key not in self.node_stds:
                continue
            if graph.ndata[key].dtype != torch.float32:
                continue
            graph.ndata[key] = (
                graph.ndata[key]
                * torch.tensor(self.node_stds[key], device=graph.device)
            ) + torch.tensor(self.node_means[key], device=graph.device)
        for key in graph.edata.keys():
            if key not in self.edge_means or key not in self.edge_stds:
                continue
            if graph.edata[key].dtype != torch.float32:
                continue
            graph.edata[key] = (
                graph.edata[key]
                * torch.tensor(self.edge_stds[key], device=graph.device)
            ) + torch.tensor(self.edge_means[key], device=graph.device)

    def denormalize(self, graph: dgl.DGLGraph) -> dgl.DGLGraph:
        """
        Denormalize the features of the graph.

        Parameters:
        -----------
        graph: dgl.DGLGraph
            The graph to denormalize.

        Returns:
        --------
        dgl.DGLGraph: The denormalized graph.
        """
        graph = graph.clone()
        self.denormalize_inplace(graph)
        return graph

    def dgl_to_pyvista_polydata(self, graph: dgl.DGLGraph) -> pv.PolyData:
        """
        Convert a DGLGraph of the volume flow field to a Pyvista PolyData object. This will not be the original mesh since the graph only contains the cell centers and the connectivity, not the cell points

        Parameters:
        -----------
        graph: dgl.DGLGraph
            The DGLGraph to convert.

        Returns:
        --------
        pv.PolyData: The converted PolyData object.
        """
        if self.do_normalization:
            graph = self.denormalize(graph)
        graph = graph.to("cpu")
        e_from, e_to = graph.edges()
        num_nodes_per_edge = torch.ones_like(e_from) * 2
        edges = torch.stack([num_nodes_per_edge, e_from, e_to], dim=1).detach()
        grid = pv.PolyData(graph.ndata["Position"].detach().numpy(), edges.numpy())
        grid.point_data["Pressure"] = graph.ndata["Pressure"].detach().numpy()
        grid.point_data["WallShearStress_0"] = (
            graph.ndata["ShearStress"][:, 0].detach().numpy()
        )
        grid.point_data["WallShearStress_1"] = (
            graph.ndata["ShearStress"][:, 1].detach().numpy()
        )
        grid.point_data["WallShearStress_2"] = (
            graph.ndata["ShearStress"][:, 2].detach().numpy()
        )
        grid.point_data["Temperature"] = graph.ndata["Temperature"].detach().numpy()
        grid.point_data["Normal_0"] = graph.ndata["Normal"][:, 0].detach().numpy()
        grid.point_data["Normal_1"] = graph.ndata["Normal"][:, 1].detach().numpy()
        grid.point_data["Normal_2"] = graph.ndata["Normal"][:, 2].detach().numpy()
        grid.point_data["CellArea"] = graph.ndata["CellArea"].detach().numpy()
        grid.point_data["BodyID"] = graph.ndata["BodyID"].detach().numpy()
        return grid

    def plot_surface(
        self,
        graph: dgl.DGLGraph,
        scalar: SurfaceFieldType,
        object_id: Optional[int] = None,
    ):
        """
        Plot the surface of the graph.

        Parameters:
        -----------
        graph: dgl.DGLGraph
            The graph to plot.
        """
        if object_id is not None:
            mask = graph.ndata["BodyID"] == object_id
            graph = graph.subgraph(mask)
        grid = self.dgl_to_pyvista_polydata(graph)
        grid.plot(
            scalars=scalar,
            render_points_as_spheres=True,
            eye_dome_lighting=True,
            render_lines_as_tubes=True,
            line_width=4,
            style="wireframe",
        )

    def compute_aggregate_force(
        self, graph: dgl.DGLGraph, object_id: Optional[int] = None
    ) -> torch.Tensor:
        """
        Compute the aggregate force acting on the surface.

        Parameters:
        -----------
        graph: dgl.DGLGraph
            The graph to compute the aggregate force for.
        object_id: Optional[int]
            The object id to compute the aggregate force for. If None, the aggregate force for all objects is computed.

        Returns:
        --------
        torch.Tensor: The aggregate force acting on the surface.
        """
        if self.do_normalization:
            graph = self.denormalize(graph)
        cell_areas = graph.ndata["CellArea"]

        cell_pressure_forces = (
            graph.ndata["Pressure"][:, None]
            * cell_areas[:, None]
            * graph.ndata["Normal"]
        )
        cell_shear_forces = graph.ndata["ShearStress"] * cell_areas[:, None]
        if object_id is not None:
            mask = graph.ndata["BodyID"] == object_id
            cell_pressure_forces = cell_pressure_forces[mask]
            cell_shear_forces = cell_shear_forces[mask]
        pressure_force = cell_pressure_forces.sum(dim=0)
        shear_force = cell_shear_forces.sum(dim=0)
        return pressure_force + shear_force

    @classmethod
    def l2_loss(cls, graph1: dgl.DGLGraph, graph2: dgl.DGLGraph):
        """
        Compute the L2 loss between two DGLGraphs.

        Parameters:
        -----------
        graph1: dgl.DGLGraph
            The first graph.
        graph2: dgl.DGLGraph
            The second graph.

        Returns:
        --------
        torch.tensor: The L2 loss between the two graphs.
        """
        # check that the graphs have the same number of nodes
        assert graph1.num_nodes() == graph2.num_nodes()
        # check that the graphs have the same number of edges
        assert graph1.num_edges() == graph2.num_edges()

        return (
            torch.nn.functional.mse_loss(
                graph1.ndata["Pressure"], graph2.ndata["Pressure"]
            )
            + torch.nn.functional.mse_loss(
                graph1.ndata["Temperature"], graph2.ndata["Temperature"]
            )
            + torch.nn.functional.mse_loss(
                graph1.ndata["ShearStress"], graph2.ndata["ShearStress"]
            )
        )



from typing import List, NamedTuple

import pandas as pd


class Position(NamedTuple):
    """Represents a 3D position. Is always the center of a body."""

    x: float
    y: float
    z: float


class Quader:
    """Represents a quader (rectangular prism) with its properties."""

    def __init__(
        self,
        temperature: float,
        position: Position,
        size_x: float,
        size_y: float,
        size_z: float,
    ):
        self.temperature = temperature
        self.position = position
        self.size_x = size_x
        self.size_y = size_y
        self.size_z = size_z

    def __repr__(self) -> str:
        return f"Quader(T={self.temperature}, Pos={self.position}, Size=({self.size_x},{self.size_y},{self.size_z}))"


class Cylinder:
    """Represents a cylinder with its properties."""

    def __init__(
        self, temperature: float, position: Position, radius: float, height: float
    ):
        self.temperature = temperature
        self.position = position
        self.radius = radius
        self.height = height

    def __repr__(self) -> str:
        return f"Cylinder(T={self.temperature}, Pos={self.position}, R={self.radius}, H={self.height})"


class SystemParameters:
    """Holds the parameters for a system of quaders, cylinders, and inflow velocity."""

    def __init__(
        self, quads: List[Quader], cylinders: List[Cylinder], inflow_velocity: float
    ):
        self.quads = quads
        self.cylinders = cylinders
        self.inflow_velocity = inflow_velocity

    def __repr__(self) -> str:
        return f"SystemParameters(V={self.inflow_velocity}, Quads={self.quads}, Cylinders={self.cylinders})"

    @staticmethod
    def from_dataframe_row(row: pd.Series) -> "SystemParameters":
        """
        Creates a SystemParameters object from a pandas DataFrame row.

        The DataFrame row is expected to contain columns as per the original CSV structure:
        T1-T6, x1-x6, y1-y6, xs1-xs4 (quad x-sizes), ys1-ys4 (quad y-sizes),
        zs1-zs4 (quad z-sizes), r5-r6 (cylinder radii), zs5-zs6 (cylinder heights), V.

        Note on Z-Coordinates for Positions:
        This method assumes that z-coordinates for the *positions* of the bodies (e.g., 'z1', 'z2', ..., 'z6')
        are NOT explicitly provided in the input DataFrame row and defaults them to 0.0.
        The fields 'zs1'-'zs4' are interpreted as z-axis *sizes* for quaders,
        and 'zs5'-'zs6' as *heights* for cylinders.
        If your DataFrame includes separate columns for positional z-coordinates (e.g. 'z1', 'z2'),
        you should modify the Position creation logic below to use `float(row[f'z{i}'])`.
        """
        quads: List[Quader] = []
        # Quaders are bodies 1, 2, 3, 4
        for i in range(1, 5):
            pos = Position(
                x=float(row[f"x{i}"]),
                y=float(row[f"y{i}"]),
                z=float(
                    row.get(f"z{i}", 0.0)
                ),  # Assumes z-location 'z{i}' might not exist, defaults to 0.0
            )
            q = Quader(
                temperature=float(row[f"T{i}"]),
                position=pos,
                size_x=float(row[f"xs{i}"]),  # xs1, xs2, xs3, xs4
                size_y=float(row[f"ys{i}"]),  # ys1, ys2, ys3, ys4
                size_z=float(
                    row[f"zs{i}"]
                ),  # zs1, zs2, zs3, zs4 (these are quad z-sizes)
            )
            quads.append(q)

        cylinders: List[Cylinder] = []
        # Cylinders are bodies 5, 6
        for i in range(5, 7):
            pos = Position(
                x=float(row[f"x{i}"]),
                y=float(row[f"y{i}"]),
                z=float(
                    row.get(f"z{i}", 0.0)
                ),  # Assumes z-location 'z{i}' might not exist, defaults to 0.0
            )
            # For cylinders, radius parameters are r5, r6 and height parameters are zs5, zs6
            cyl = Cylinder(
                temperature=float(row[f"T{i}"]),
                position=pos,
                radius=float(row[f"r{i}"]),  # r5, r6
                height=float(row[f"zs{i}"]),  # zs5, zs6 (these are cylinder heights)
            )
            cylinders.append(cyl)

        inflow_velocity = float(row["V"])

        return SystemParameters(
            quads=quads, cylinders=cylinders, inflow_velocity=inflow_velocity
        )


def df_row_to_system_parameters(df: pd.DataFrame, design_id: int) -> SystemParameters:
    """
    Helper function to create a SystemParameters object from a specific row of a pandas DataFrame.
    """
    if not isinstance(df, pd.DataFrame):
        raise TypeError("Input 'df' must be a pandas DataFrame.")
    if not isinstance(design_id, int):
        raise TypeError("'design_id' must be an integer.")
    row_id = design_id - 1
    if not 0 <= row_id < len(df):
        raise IndexError(
            f"'row_id' {row_id} is out of bounds for DataFrame with {len(df)} rows."
        )

    row_series = df.iloc[row_id]
    return SystemParameters.from_dataframe_row(row_series)



import os
import re
import shutil
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
from pathlib import Path
from typing import Literal

import huggingface_hub as hf  # type: ignore
import numpy as np
import pandas as pd
import pyvista as pv
from tqdm import tqdm

from cooldata.metadata import SystemParameters, df_row_to_system_parameters

VolumeFieldType = Literal[
    "Velocity_0",
    "Velocity_1",
    "Velocity_2",
    "Pressure",
    "Temperature",
    "TurbulentKineticEnergy",
    "TurbulentDissipationRate",
]


SurfaceFieldType = Literal[
    "AreaMagnitude",
    "HeatTransferCoefficient",
    "Normal_0",
    "Normal_1",
    "Normal_2",
    "Temperature",
    "Pressure",
    "WallShearStressMagnitude",
    "WallShearStress_0",
    "WallShearStress_1",
    "WallShearStress_2",
]


class PyvistaSample:
    """
    A PyVista sample representing a 3D flow field dataset. The data is stored on disk and can be loaded to a PyVista object when needed
    """
    def __init__(
        self,
        volume_path: str | Path,
        surface_path: str | Path,
        metadata: SystemParameters | None = None,
    ):
        self.volume_path = Path(volume_path)
        self.surface_path = Path(surface_path)
        self._volume_data: pv.MultiBlock | None = None
        self._surface_data: pv.MultiBlock | None = None
        self.metadata = metadata
        self._bounds: tuple[float, float, float, float, float, float] | None = None

    @property
    def surface_data(self):
        if self._surface_data is None:
            self._surface_data = pv.read(self.surface_path)
        # TODO: Check if data is a valid flow field dataset, i.e., has the necessary point data
        return self._surface_data

    @property
    def volume_data(self):
        if self._volume_data is None:
            self._volume_data = pv.read(self.volume_path)
        # TODO: Check if data is a valid flow field dataset, i.e., has the necessary point data
        return self._volume_data

    def plot_surface(self, field: SurfaceFieldType):
        self.surface_data.plot(scalars=field)

    def plot_volume(self, field: VolumeFieldType):
        self.volume_data[0][0][0].plot(scalars=field, opacity=0.7)

    def get_points(self) -> np.ndarray:
        """
        Returns the points of the dataset as a numpy array.

        Returns:
        --------
        np.ndarray: The points of the dataset. Shape: (n_points, 3)
        """
        # TODO check if same works for non cgns
        block = self.volume_data[0][0][0]
        return block.cell_centers().points

    def get_surface_points(self, block_index: int) -> np.ndarray:
        """
        Returns the points of the surface dataset as a numpy array.

        Returns:
        --------
        np.ndarray: The points of the surface dataset. Shape: (n_points, 3)
        """
        block = self.surface_data[0][block_index]
        return block.points

    def get_labeled_surface_points(self) -> np.ndarray:
        """
        Returns the surface points of the dataset with their block index as a numpy array.

        Returns:
        --------
        np.ndarray: The labeled points of the dataset. Shape: (n_points, 4)
        """
        labeled_points = []
        for i, block in enumerate(self.surface_data[0]):
            labeled_points.append(
                np.hstack((block.points, np.full((block.n_points, 1), i)))
            )
        return np.vstack(labeled_points)

    @property
    def is_loaded(self) -> bool:
        """
        Returns True if the volume and surface data are loaded, False otherwise.
        """
        return self._volume_data is not None and self._surface_data is not None

    def load(self):
        self.volume_data
        self.surface_data

    def unload(self):
        # TODO: Test if this frees the memory
        self._volume_data = None
        self._surface_data = None

    def get_bounding_box(self):
        """
        Returns the bounding box of the volume data.
        The bounding box is a six-tuple (xmin, xmax, ymin, ymax, zmin, zmax).
        """
        if self._bounds is not None:
            return self._bounds
        was_loaded = self.is_loaded
        self._bounds = self.volume_data.bounds
        if not was_loaded:
            self.unload()
        return self._bounds

    @property
    def design_id(self) -> int:
        """
        Returns the design ID of the sample, which is extracted from the file name.
        The design ID is assumed to be the second last part of the file name, split by underscores.
        """
        stem = self.volume_path.stem
        id = re.search(r"_(\d+)", stem)
        if id:
            return int(id.group(1))
        raise ValueError("Design ID not found")

    def _repr_html_(self) -> str:
        if not hasattr(self, "metadata") or self.metadata is None:
            return "<p>No metadata available for this sample to generate HTML representation.</p>"

        # Ensure metadata object is not None and has quads and cylinders attributes
        if not hasattr(self.metadata, "quads") or not hasattr(
            self.metadata, "cylinders"
        ):
            return "<p>Metadata object is malformed (missing quads or cylinders attributes).</p>"

        _design_id_str: str
        try:
            _design_id_str = str(self.design_id)
        except ValueError:
            _design_id_str = "N_A"  # Use a safe fallback for ID

        try:
            bounds = self.get_bounding_box()  # (xmin, xmax, ymin, ymax, zmin, zmax)
        except Exception as e:
            return f"""<p>Error getting bounding box for sample ID {_design_id_str}: {e}. 
Make sure volume data is accessible and valid.</p>"""

        xmin, xmax, ymin, ymax, _, _ = bounds

        world_width = xmax - xmin
        world_height = ymax - ymin

        if (
            world_width <= 1e-6 or world_height <= 1e-6
        ):  # Using a small epsilon for float comparison
            return f"""<p>Bounding box for sample ID {_design_id_str} has zero or negligible 2D dimensions 
(width: {world_width:.2e}, height: {world_height:.2e}). Cannot render.</p>"""

        # SVG viewport dimensions (fixed size for the output image)
        svg_width = 300
        svg_height = 200

        padding = 0.05 * max(world_width, world_height)

        view_box_xmin = xmin - padding
        view_box_ymin = ymin - padding
        view_box_width = world_width + 2 * padding
        view_box_height = world_height + 2 * padding

        # Determine a reasonable stroke width based on world dimensions
        # This aims for a stroke that's roughly 0.5% of the main dimension
        dynamic_stroke_width = 0.005 * max(world_width, world_height)

        html_parts = [
            '<div style="font-family: Arial, sans-serif; margin: 10px; padding: 10px; border: 1px solid #e0e0e0; border-radius: 5px; background-color: #fdfdfd;">',
            f'  <h4 style="margin-top:0; margin-bottom: 10px; color: #333;">Design ID: {_design_id_str}</h4>',
            f'  <svg width="{svg_width}" height="{svg_height}" viewBox="{view_box_xmin} {view_box_ymin} {view_box_width} {view_box_height}" style="border:1px solid #ccc; background-color: #ffffff;">',
        ]

        # Unique ID for clipPath, incorporating object id for robustness in some edge cases
        clip_path_id = f"clipPath_PyvistaSample_{_design_id_str}_{hex(id(self))[-6:]}"

        html_parts.append("    <defs>")
        html_parts.append(f'      <clipPath id="{clip_path_id}">')
        html_parts.append(
            f'        <rect x="{xmin}" y="{ymin}" width="{world_width}" height="{world_height}" />'
        )
        html_parts.append("      </clipPath>")
        html_parts.append("    </defs>")

        html_parts.append(
            f'    <rect x="{xmin}" y="{ymin}" width="{world_width}" height="{world_height}" fill="none" stroke="#bbbbbb" stroke-width="{dynamic_stroke_width}" stroke-dasharray="{2 * dynamic_stroke_width},{2 * dynamic_stroke_width}" />'
        )

        html_parts.append(f'    <g clip-path="url(#{clip_path_id})">')

        quad_colors = ["#1f77b4", "#ff7f0e", "#2ca02c", "#d62728", "#9467bd", "#8c564b"]
        for i, quader in enumerate(self.metadata.quads):
            q_x = quader.position.x - quader.size_x / 2
            q_y = quader.position.y - quader.size_y / 2
            color = quad_colors[i % len(quad_colors)]
            html_parts.append(
                f'      <rect x="{q_x}" y="{q_y}" width="{quader.size_x}" height="{quader.size_y}" fill="{color}" fill-opacity="0.6" stroke="{color}" stroke-opacity="0.9" stroke-width="{0.5 * dynamic_stroke_width}">'
            )
            html_parts.append(
                f"        <title>Quader {i + 1}\\nTemp: {quader.temperature:.1f}\\nPos: ({quader.position.x:.2f}, {quader.position.y:.2f})\\nSize: ({quader.size_x:.2f}, {quader.size_y:.2f})</title>"
            )
            html_parts.append("      </rect>")

        cyl_colors = ["#17becf", "#bcbd22", "#e377c2", "#7f7f7f", "#aec7e8", "#ffbb78"]
        for i, cylinder in enumerate(self.metadata.cylinders):
            color = cyl_colors[i % len(cyl_colors)]
            html_parts.append(
                f'      <circle cx="{cylinder.position.x}" cy="{cylinder.position.y}" r="{cylinder.radius}" fill="{color}" fill-opacity="0.6" stroke="{color}" stroke-opacity="0.9" stroke-width="{0.5 * dynamic_stroke_width}">'
            )
            html_parts.append(
                f"        <title>Cylinder {i + 1}\\nTemp: {cylinder.temperature:.1f}\\nPos: ({cylinder.position.x:.2f}, {cylinder.position.y:.2f})\\nRadius: {cylinder.radius:.2f}</title>"
            )
            html_parts.append("      </circle>")

        html_parts.append("    </g>")
        html_parts.append("  </svg>")

        html_parts.append(
            f'  <p style="font-size: 0.8em; color: #666; margin-top: 8px; margin-bottom: 0;">Bounding Box (xmin, xmax, ymin, ymax): ({xmin:.2f}, {xmax:.2f}, {ymin:.2f}, {ymax:.2f})</p>'
        )
        html_parts.append("</div>")

        return "\n".join(html_parts)


class PyvistaFlowFieldDataset:
    """
    The main class for working with the cooldata dataset.
    """
    def __init__(self, samples: list[PyvistaSample]):
        self.samples = samples

    def __len__(self):
        return len(self.samples)

    def __getitem__(self, idx: int):
        return self.samples[idx]

    def slice(self, start: int, end: int):
        """
        Returns a slice of the dataset.
        """
        return PyvistaFlowFieldDataset(self.samples[start:end])

    def shuffle(self):
        """
        Shuffles the dataset in place.
        """
        np.random.shuffle(self.samples)

    @classmethod
    def try_from_directory(
        cls, data_dir: str | Path, num_samples: int
    ) -> "None | PyvistaFlowFieldDataset":
        data_dir = os.path.abspath(data_dir)
        data_dir = Path(data_dir)
        volume_dir = data_dir / "volume"
        surface_dir = data_dir / "surface"
        metadata_file = data_dir / "metadata.parquet"
        if not metadata_file.exists():
            print(f"Metadata file not found at {metadata_file}.")
            return None
        os.makedirs(volume_dir, exist_ok=True)
        os.makedirs(surface_dir, exist_ok=True)
        volume_files = list(volume_dir.glob("*.cgns"))
        surface_files = list(surface_dir.glob("*.cgns"))
        volume_indices = [int(f.stem.split("_")[-2]) for f in volume_files]
        surface_indices = [int(f.stem.split("_")[-2]) for f in surface_files]
        volume_indices.sort()
        surface_indices.sort()
        if volume_indices != surface_indices:
            return None
        volume_files = sorted(volume_files, key=lambda x: int(x.stem.split("_")[-2]))
        surface_files = sorted(surface_files, key=lambda x: int(x.stem.split("_")[-2]))
        samples = [PyvistaSample(v, s) for v, s in zip(volume_files, surface_files)]
        if len(samples) < num_samples:
            return None
        samples = samples[:num_samples]
        ds = cls(samples)
        metadata_df = pd.read_parquet(metadata_file)
        ds.add_metadata(metadata_df)
        print(f"Loaded {len(ds)} samples from '{data_dir}'.")
        return ds

    @classmethod
    def load_from_huggingface(
        cls, data_dir: str | Path, num_samples=3
    ) -> "PyvistaFlowFieldDataset":
        """
        Download the given number of samples from huggingface to data_dir.
        Args:
        - data_dir: The directory to download the data to.
        - num_samples: The number of samples to download
        """
        loaded = cls.try_from_directory(data_dir, num_samples)
        if loaded is not None:
            print(f"Loaded {len(loaded)} samples from '{data_dir}'.")
            return loaded

        data_dir = os.path.abspath(data_dir)
        data_dir = Path(data_dir)
        volume_dir = data_dir / "volume"
        surface_dir = data_dir / "surface"
        tmp_dir = data_dir / "tmp"
        os.makedirs(tmp_dir, exist_ok=True)

        # remove existing files
        if os.path.exists(volume_dir):
            shutil.rmtree(volume_dir)
        if os.path.exists(surface_dir):
            shutil.rmtree(surface_dir)
        os.makedirs(volume_dir, exist_ok=True)
        os.makedirs(surface_dir, exist_ok=True)

        repo_id = "datasets/bgce/cooldata-v2"
        fs = hf.HfFileSystem()
        # download metadata file
        metadata_file = f"{repo_id}/metadata.parquet"
        local_metadata_path = os.path.join(data_dir, "metadata.parquet")
        fs.download(metadata_file, local_metadata_path)
        metadata_df = pd.read_parquet(local_metadata_path)

        runs = fs.glob(f"{repo_id}/runs/run_*", detail=False)
        samples: list[PyvistaSample] = []
        runs = sorted(runs, key=lambda x: int(x.split("/")[-1].removeprefix("run_")))
        for run in runs:
            run_name = os.path.basename(run)
            zip_files_in_run = list(fs.glob(f"{run}/batch_*.zip", detail=False))
            zip_files_in_run = sorted(
                zip_files_in_run,
                key=lambda x: int(
                    x.split("/")[-1].removeprefix("batch_").removesuffix(".zip")
                ),
            )
            for zip_file in zip_files_in_run:
                local_path = os.path.join(tmp_dir, run_name, os.path.basename(zip_file))
                try:
                    fs.download(zip_file, local_path)
                except Exception as e:
                    print(f"Failed to download {zip_file} for run {run_name}: {e}")
                    continue
                # Extract the zip file
                unzip_dir = os.path.join(
                    tmp_dir, run_name, os.path.basename(zip_file).removesuffix(".zip")
                )
                os.makedirs(unzip_dir, exist_ok=True)
                shutil.unpack_archive(local_path, unzip_dir)
                # Match indices
                volume_files = list(Path(unzip_dir).glob("volume_design_*_p.cgns"))
                surface_files = list(Path(unzip_dir).glob("surface_design_*_p.cgns"))
                volume_indices = [int(f.stem.split("_")[-2]) for f in volume_files]
                surface_indices = [int(f.stem.split("_")[-2]) for f in surface_files]
                volume_indices.sort()
                surface_indices.sort()
                if volume_indices != surface_indices:
                    print(f"Skipping {run_name} due to mismatched indices.")
                    continue
                volume_files = sorted(
                    volume_files, key=lambda x: int(x.stem.split("_")[-2])
                )
                surface_files = sorted(
                    surface_files, key=lambda x: int(x.stem.split("_")[-2])
                )
                for v, s in zip(volume_files, surface_files):
                    # Copy the files to the new directory
                    shutil.copy(v, volume_dir)
                    shutil.copy(s, surface_dir)
                    moved_volume_path = volume_dir / v.name
                    moved_surface_path = surface_dir / s.name
                    samples.append(PyvistaSample(moved_volume_path, moved_surface_path))
                    if len(samples) >= num_samples:
                        shutil.rmtree(tmp_dir)
                        ds = cls(samples)
                        ds.add_metadata(metadata_df)
                        print(f"Loaded {len(ds)} samples from '{data_dir}'.")
                        return ds
        # Clean up temporary directory
        shutil.rmtree(tmp_dir)

        ds = cls(samples)
        ds.add_metadata(metadata_df)
        print(f"Loaded {len(ds)} samples from '{data_dir}'.")
        return ds

    def add_metadata(self, metadata_df: pd.DataFrame) -> None:
        """
        Adds metadata to the samples from a pandas DataFrame.
        The DataFrame should have a column 'design_id' that matches the design ID of the samples.
        """
        if not isinstance(metadata_df, pd.DataFrame):
            raise TypeError("metadata_df must be a pandas DataFrame.")

        for sample in self.samples:
            design_id = sample.design_id
            try:
                md = df_row_to_system_parameters(metadata_df, design_id)
                sample.metadata = md
            except Exception as e:
                print(
                    f"Failed to add metadata for sample with design ID {design_id}: {e}"
                )

    def get_bounds(self):
        """
        Returns the bounding box of the volume data.
        The bounding box is a six-tuple (xmin, xmax, ymin, ymax, zmin, zmax).
        """
        bounds = (np.inf, -np.inf, np.inf, -np.inf, np.inf, -np.inf)
        self.unload()

        with ProcessPoolExecutor() as executor:
            sample_bounds_list = list(
                tqdm(
                    executor.map(get_sample_bounds, self.samples),
                    total=len(self.samples),
                    desc="Computing bounds",
                )
            )

        for sample_bounds in sample_bounds_list:
            bounds = (
                min(bounds[0], sample_bounds[0]),
                max(bounds[1], sample_bounds[1]),
                min(bounds[2], sample_bounds[2]),
                max(bounds[3], sample_bounds[3]),
                min(bounds[4], sample_bounds[4]),
                max(bounds[5], sample_bounds[5]),
            )

        return bounds

    def load_to_memory(self):
        """
        Load all samples into memory.
        """
        for sample in self.samples:
            sample.load()

    def unload(self):
        """
        Unload all samples from memory.
        """
        for sample in self.samples:
            sample.unload()


def get_sample_bounds(sample):
    return sample.get_bounding_box()
